/* caDIMClientRecord.c */
/* Example record support module */
  
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "alarm.h"
#include "dbAccess.h"
#include "recGbl.h"
#include "dbEvent.h"
#include "dbDefs.h"
#include "dbAccess.h"
#include "devSup.h"
#include "errMdef.h"
#include "recSup.h"
#include "special.h"
#define GEN_SIZE_OFFSET
#include "caDIMClientRecord.h"
#undef  GEN_SIZE_OFFSET
#include "epicsExport.h"

/* Create RSET - Record Support Entry Table */
#define report NULL
#define initialize NULL
static long init_record();
static long process();
#define special NULL
#define get_value NULL
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
static long get_units();
static long get_precision();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double();
static long get_control_double();
static long get_alarm_double();
 
rset caDIMClientRSET={
	RSETNUMBER,
	report,
	initialize,
	init_record,
	process,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double
};
epicsExportAddress(rset,caDIMClientRSET);

typedef struct caDIMClientset { /* caDIMClient input dset */
	long		number;
	DEVSUPFUN	dev_report;
	DEVSUPFUN	init;
	DEVSUPFUN	init_record; /*returns: (-1,0)=>(failure,success)*/
	DEVSUPFUN	get_ioint_info;
	DEVSUPFUN	read_caDIMClient;
}caDIMClientdset;

static void checkAlarms(caDIMClientRecord *pcaDIMClient);
static void monitor(caDIMClientRecord *pcaDIMClient);

static long init_record(void *precord,int pass)
{
    caDIMClientRecord	*pcaDIMClient = (caDIMClientRecord *)precord;
    caDIMClientdset	*pdset;
    long	status;

    if (pass==0) return(0);

    if(!(pdset = (caDIMClientdset *)(pcaDIMClient->dset))) {
	recGblRecordError(S_dev_noDSET,(void *)pcaDIMClient,"caDIMClient: init_record");
	return(S_dev_noDSET);
    }
    /* must have read_caDIMClient function defined */
    if( (pdset->number < 5) || (pdset->read_caDIMClient == NULL) ) {
	recGblRecordError(S_dev_missingSup,(void *)pcaDIMClient,"caDIMClient: init_record");
	return(S_dev_missingSup);
    }

    if( pdset->init_record ) {
	if((status=(*pdset->init_record)(pcaDIMClient))) return(status);
    }
    return(0);
}

static long process(void *precord)
{
	caDIMClientRecord	*pcaDIMClient = (caDIMClientRecord *)precord;
	caDIMClientdset		*pdset = (caDIMClientdset *)(pcaDIMClient->dset);
	long		 status;
	unsigned char    pact=pcaDIMClient->pact;

	if( (pdset==NULL) || (pdset->read_caDIMClient==NULL) ) {
		pcaDIMClient->pact=TRUE;
		recGblRecordError(S_dev_missingSup,(void *)pcaDIMClient,"read_caDIMClient");
		return(S_dev_missingSup);
	}

	/* pact must not be set until after calling device support */
	status=(*pdset->read_caDIMClient)(pcaDIMClient);
	/* check if device support set pact */
	if ( !pact && pcaDIMClient->pact ) return(0);
	pcaDIMClient->pact = TRUE;

	recGblGetTimeStamp(pcaDIMClient);
	/* check for alarms */
	checkAlarms(pcaDIMClient);
	/* check event list */
	monitor(pcaDIMClient);
	/* process the forward scan link record */
        recGblFwdLink(pcaDIMClient);

	pcaDIMClient->pact=FALSE;
	return(status);
}

static long get_units(DBADDR *paddr, char *units)
{
    caDIMClientRecord	*pcaDIMClient=(caDIMClientRecord *)paddr->precord;

    strncpy(units,pcaDIMClient->egu,DB_UNITS_SIZE);
    return(0);
}

static long get_precision(DBADDR *paddr, long *precision)
{
    caDIMClientRecord	*pcaDIMClient=(caDIMClientRecord *)paddr->precord;

    *precision = pcaDIMClient->prec;
    if(paddr->pfield == (void *)&pcaDIMClient->val) return(0);
    recGblGetPrec(paddr,precision);
    return(0);
}

static long get_graphic_double(DBADDR *paddr,struct dbr_grDouble *pgd)
{
    caDIMClientRecord	*pcaDIMClient=(caDIMClientRecord *)paddr->precord;
    int		fieldIndex = dbGetFieldIndex(paddr);

    if(fieldIndex == caDIMClientRecordVAL
    || fieldIndex == caDIMClientRecordHIHI
    || fieldIndex == caDIMClientRecordHIGH
    || fieldIndex == caDIMClientRecordLOW
    || fieldIndex == caDIMClientRecordLOLO
    || fieldIndex == caDIMClientRecordHOPR
    || fieldIndex == caDIMClientRecordLOPR) {
        pgd->upper_disp_limit = pcaDIMClient->hopr;
        pgd->lower_disp_limit = pcaDIMClient->lopr;
    } else recGblGetGraphicDouble(paddr,pgd);
    return(0);
}

static long get_control_double(DBADDR *paddr,struct dbr_ctrlDouble *pcd)
{
    caDIMClientRecord	*pcaDIMClient=(caDIMClientRecord *)paddr->precord;
    int		fieldIndex = dbGetFieldIndex(paddr);

    if(fieldIndex == caDIMClientRecordVAL
    || fieldIndex == caDIMClientRecordHIHI
    || fieldIndex == caDIMClientRecordHIGH
    || fieldIndex == caDIMClientRecordLOW
    || fieldIndex == caDIMClientRecordLOLO) {
	pcd->upper_ctrl_limit = pcaDIMClient->hopr;
	pcd->lower_ctrl_limit = pcaDIMClient->lopr;
    } else recGblGetControlDouble(paddr,pcd);
    return(0);
}

static long get_alarm_double(DBADDR *paddr,struct dbr_alDouble *pad)
{
    caDIMClientRecord	*pcaDIMClient=(caDIMClientRecord *)paddr->precord;
    int		fieldIndex = dbGetFieldIndex(paddr);

    if(fieldIndex == caDIMClientRecordVAL) {
	pad->upper_alarm_limit = pcaDIMClient->hihi;
	pad->upper_warning_limit = pcaDIMClient->high;
	pad->lower_warning_limit = pcaDIMClient->low;
	pad->lower_alarm_limit = pcaDIMClient->lolo;
    } else recGblGetAlarmDouble(paddr,pad);
    return(0);
}

static void checkAlarms(caDIMClientRecord *pcaDIMClient)
{
	double		val;
	float		hyst, lalm, hihi, high, low, lolo;
	unsigned short	hhsv, llsv, hsv, lsv;

	if(pcaDIMClient->udf == TRUE ){
		recGblSetSevr(pcaDIMClient,UDF_ALARM,INVALID_ALARM);
		return;
	}
	hihi = pcaDIMClient->hihi; lolo = pcaDIMClient->lolo; high = pcaDIMClient->high; low = pcaDIMClient->low;
	hhsv = pcaDIMClient->hhsv; llsv = pcaDIMClient->llsv; hsv = pcaDIMClient->hsv; lsv = pcaDIMClient->lsv;
	val = pcaDIMClient->val; hyst = pcaDIMClient->hyst; lalm = pcaDIMClient->lalm;

	/* alarm condition hihi */
	if (hhsv && (val >= hihi || ((lalm==hihi) && (val >= hihi-hyst)))){
	        if (recGblSetSevr(pcaDIMClient,HIHI_ALARM,pcaDIMClient->hhsv)) pcaDIMClient->lalm = hihi;
		return;
	}

	/* alarm condition lolo */
	if (llsv && (val <= lolo || ((lalm==lolo) && (val <= lolo+hyst)))){
	        if (recGblSetSevr(pcaDIMClient,LOLO_ALARM,pcaDIMClient->llsv)) pcaDIMClient->lalm = lolo;
		return;
	}

	/* alarm condition high */
	if (hsv && (val >= high || ((lalm==high) && (val >= high-hyst)))){
	        if (recGblSetSevr(pcaDIMClient,HIGH_ALARM,pcaDIMClient->hsv)) pcaDIMClient->lalm = high;
		return;
	}

	/* alarm condition low */
	if (lsv && (val <= low || ((lalm==low) && (val <= low+hyst)))){
	        if (recGblSetSevr(pcaDIMClient,LOW_ALARM,pcaDIMClient->lsv)) pcaDIMClient->lalm = low;
		return;
	}

	/* we get here only if val is out of alarm by at least hyst */
	pcaDIMClient->lalm = val;
	return;
}

static void monitor(caDIMClientRecord *pcaDIMClient)
{
	unsigned short	monitor_mask;
	double		delta;

        monitor_mask = recGblResetAlarms(pcaDIMClient);
	/* check for value change */
	delta = pcaDIMClient->mlst - pcaDIMClient->val;
	if(delta<0.0) delta = -delta;
	if (delta > pcaDIMClient->mdel) {
		/* post events for value change */
		monitor_mask |= DBE_VALUE;
		/* update last value monitored */
		pcaDIMClient->mlst = pcaDIMClient->val;
	}

	/* check for archive change */
	delta = pcaDIMClient->alst - pcaDIMClient->val;
	if(delta<0.0) delta = -delta;
	if (delta > pcaDIMClient->adel) {
		/* post events on value field for archive change */
		monitor_mask |= DBE_LOG;
		/* update last archive value monitored */
		pcaDIMClient->alst = pcaDIMClient->val;
	}

	/* send out monitors connected to the value field */
	if (monitor_mask){
		db_post_events(pcaDIMClient,&pcaDIMClient->val,monitor_mask);
	}
	return;
}
