/* caDIMServerRecord.c */
/* Example record support module */
  
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "alarm.h"
#include "dbAccess.h"
#include "recGbl.h"
#include "dbEvent.h"
#include "dbDefs.h"
#include "dbAccess.h"
#include "devSup.h"
#include "errMdef.h"
#include "recSup.h"
#include "special.h"
#define GEN_SIZE_OFFSET
#include "caDIMServerRecord.h"
#undef  GEN_SIZE_OFFSET
#include "epicsExport.h"

/* Create RSET - Record Support Entry Table */
#define report NULL
#define initialize NULL
static long init_record();
static long process();
#define special NULL
#define get_value NULL
#define cvt_dbaddr NULL
#define get_array_info NULL
#define put_array_info NULL
static long get_units();
static long get_precision();
#define get_enum_str NULL
#define get_enum_strs NULL
#define put_enum_str NULL
static long get_graphic_double();
static long get_control_double();
static long get_alarm_double();
 
rset caDIMServerRSET={
	RSETNUMBER,
	report,
	initialize,
	init_record,
	process,
	special,
	get_value,
	cvt_dbaddr,
	get_array_info,
	put_array_info,
	get_units,
	get_precision,
	get_enum_str,
	get_enum_strs,
	put_enum_str,
	get_graphic_double,
	get_control_double,
	get_alarm_double
};
epicsExportAddress(rset,caDIMServerRSET);

typedef struct caDIMServerset { /* caDIMServer input dset */
	long		number;
	DEVSUPFUN	dev_report;
	DEVSUPFUN	init;
	DEVSUPFUN	init_record; /*returns: (-1,0)=>(failure,success)*/
	DEVSUPFUN	get_ioint_info;
	DEVSUPFUN	read_caDIMServer;
}caDIMServerdset;

static void checkAlarms(caDIMServerRecord *pcaDIMServer);
static void monitor(caDIMServerRecord *pcaDIMServer);

static long init_record(void *precord,int pass)
{
    caDIMServerRecord	*pcaDIMServer = (caDIMServerRecord *)precord;
    caDIMServerdset	*pdset;
    long	status;

    if (pass==0) return(0);

    if(!(pdset = (caDIMServerdset *)(pcaDIMServer->dset))) {
	recGblRecordError(S_dev_noDSET,(void *)pcaDIMServer,"caDIMServer: init_record");
	return(S_dev_noDSET);
    }
    /* must have read_caDIMServer function defined */
    if( (pdset->number < 5) || (pdset->read_caDIMServer == NULL) ) {
	recGblRecordError(S_dev_missingSup,(void *)pcaDIMServer,"caDIMServer: init_record");
	return(S_dev_missingSup);
    }

    if( pdset->init_record ) {
	if((status=(*pdset->init_record)(pcaDIMServer))) return(status);
    }
    return(0);
}

static long process(void *precord)
{
	caDIMServerRecord	*pcaDIMServer = (caDIMServerRecord *)precord;
	caDIMServerdset		*pdset = (caDIMServerdset *)(pcaDIMServer->dset);
	long		 status;
	unsigned char    pact=pcaDIMServer->pact;

	if( (pdset==NULL) || (pdset->read_caDIMServer==NULL) ) {
		pcaDIMServer->pact=TRUE;
		recGblRecordError(S_dev_missingSup,(void *)pcaDIMServer,"read_caDIMServer");
		return(S_dev_missingSup);
	}

	/* pact must not be set until after calling device support */
	status=(*pdset->read_caDIMServer)(pcaDIMServer);
	/* check if device support set pact */
	if ( !pact && pcaDIMServer->pact ) return(0);
	pcaDIMServer->pact = TRUE;

	recGblGetTimeStamp(pcaDIMServer);
	/* check for alarms */
	checkAlarms(pcaDIMServer);
	/* check event list */
	monitor(pcaDIMServer);
	/* process the forward scan link record */
        recGblFwdLink(pcaDIMServer);

	pcaDIMServer->pact=FALSE;
	return(status);
}

static long get_units(DBADDR *paddr, char *units)
{
    caDIMServerRecord	*pcaDIMServer=(caDIMServerRecord *)paddr->precord;

    strncpy(units,pcaDIMServer->egu,DB_UNITS_SIZE);
    return(0);
}

static long get_precision(DBADDR *paddr, long *precision)
{
    caDIMServerRecord	*pcaDIMServer=(caDIMServerRecord *)paddr->precord;

    *precision = pcaDIMServer->prec;
    if(paddr->pfield == (void *)&pcaDIMServer->val) return(0);
    recGblGetPrec(paddr,precision);
    return(0);
}

static long get_graphic_double(DBADDR *paddr,struct dbr_grDouble *pgd)
{
    caDIMServerRecord	*pcaDIMServer=(caDIMServerRecord *)paddr->precord;
    int		fieldIndex = dbGetFieldIndex(paddr);

    if(fieldIndex == caDIMServerRecordVAL
    || fieldIndex == caDIMServerRecordHIHI
    || fieldIndex == caDIMServerRecordHIGH
    || fieldIndex == caDIMServerRecordLOW
    || fieldIndex == caDIMServerRecordLOLO
    || fieldIndex == caDIMServerRecordHOPR
    || fieldIndex == caDIMServerRecordLOPR) {
        pgd->upper_disp_limit = pcaDIMServer->hopr;
        pgd->lower_disp_limit = pcaDIMServer->lopr;
    } else recGblGetGraphicDouble(paddr,pgd);
    return(0);
}

static long get_control_double(DBADDR *paddr,struct dbr_ctrlDouble *pcd)
{
    caDIMServerRecord	*pcaDIMServer=(caDIMServerRecord *)paddr->precord;
    int		fieldIndex = dbGetFieldIndex(paddr);

    if(fieldIndex == caDIMServerRecordVAL
    || fieldIndex == caDIMServerRecordHIHI
    || fieldIndex == caDIMServerRecordHIGH
    || fieldIndex == caDIMServerRecordLOW
    || fieldIndex == caDIMServerRecordLOLO) {
	pcd->upper_ctrl_limit = pcaDIMServer->hopr;
	pcd->lower_ctrl_limit = pcaDIMServer->lopr;
    } else recGblGetControlDouble(paddr,pcd);
    return(0);
}

static long get_alarm_double(DBADDR *paddr,struct dbr_alDouble *pad)
{
    caDIMServerRecord	*pcaDIMServer=(caDIMServerRecord *)paddr->precord;
    int		fieldIndex = dbGetFieldIndex(paddr);

    if(fieldIndex == caDIMServerRecordVAL) {
	pad->upper_alarm_limit = pcaDIMServer->hihi;
	pad->upper_warning_limit = pcaDIMServer->high;
	pad->lower_warning_limit = pcaDIMServer->low;
	pad->lower_alarm_limit = pcaDIMServer->lolo;
    } else recGblGetAlarmDouble(paddr,pad);
    return(0);
}

static void checkAlarms(caDIMServerRecord *pcaDIMServer)
{
	double		val;
	float		hyst, lalm, hihi, high, low, lolo;
	unsigned short	hhsv, llsv, hsv, lsv;

	if(pcaDIMServer->udf == TRUE ){
		recGblSetSevr(pcaDIMServer,UDF_ALARM,INVALID_ALARM);
		return;
	}
	hihi = pcaDIMServer->hihi; lolo = pcaDIMServer->lolo; high = pcaDIMServer->high; low = pcaDIMServer->low;
	hhsv = pcaDIMServer->hhsv; llsv = pcaDIMServer->llsv; hsv = pcaDIMServer->hsv; lsv = pcaDIMServer->lsv;
	val = pcaDIMServer->val; hyst = pcaDIMServer->hyst; lalm = pcaDIMServer->lalm;

	/* alarm condition hihi */
	if (hhsv && (val >= hihi || ((lalm==hihi) && (val >= hihi-hyst)))){
	        if (recGblSetSevr(pcaDIMServer,HIHI_ALARM,pcaDIMServer->hhsv)) pcaDIMServer->lalm = hihi;
		return;
	}

	/* alarm condition lolo */
	if (llsv && (val <= lolo || ((lalm==lolo) && (val <= lolo+hyst)))){
	        if (recGblSetSevr(pcaDIMServer,LOLO_ALARM,pcaDIMServer->llsv)) pcaDIMServer->lalm = lolo;
		return;
	}

	/* alarm condition high */
	if (hsv && (val >= high || ((lalm==high) && (val >= high-hyst)))){
	        if (recGblSetSevr(pcaDIMServer,HIGH_ALARM,pcaDIMServer->hsv)) pcaDIMServer->lalm = high;
		return;
	}

	/* alarm condition low */
	if (lsv && (val <= low || ((lalm==low) && (val <= low+hyst)))){
	        if (recGblSetSevr(pcaDIMServer,LOW_ALARM,pcaDIMServer->lsv)) pcaDIMServer->lalm = low;
		return;
	}

	/* we get here only if val is out of alarm by at least hyst */
	pcaDIMServer->lalm = val;
	return;
}

static void monitor(caDIMServerRecord *pcaDIMServer)
{
	unsigned short	monitor_mask;
	double		delta;

        monitor_mask = recGblResetAlarms(pcaDIMServer);
	/* check for value change */
	delta = pcaDIMServer->mlst - pcaDIMServer->val;
	if(delta<0.0) delta = -delta;
	if (delta > pcaDIMServer->mdel) {
		/* post events for value change */
		monitor_mask |= DBE_VALUE;
		/* update last value monitored */
		pcaDIMServer->mlst = pcaDIMServer->val;
	}

	/* check for archive change */
	delta = pcaDIMServer->alst - pcaDIMServer->val;
	if(delta<0.0) delta = -delta;
	if (delta > pcaDIMServer->adel) {
		/* post events on value field for archive change */
		monitor_mask |= DBE_LOG;
		/* update last archive value monitored */
		pcaDIMServer->alst = pcaDIMServer->val;
	}

	/* send out monitors connected to the value field */
	if (monitor_mask){
		db_post_events(pcaDIMServer,&pcaDIMServer->val,monitor_mask);
	}
	return;
}
