/* devcaDIMServerSoft.c */
/* Example device support module */

#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "alarm.h"
#include "cvtTable.h"
#include "dbDefs.h"
#include "dbAccess.h"
#include "recGbl.h"
#include "recSup.h"
#include "devSup.h"
#include "link.h"
#include "caDIMServerRecord.h"
#include "epicsExport.h"

/*DIM specific */
#include "dis.h"
#define EPICS_DIM_SERVER "EPICS_IOC_1"
/*DIM specific */

/*DIM declarations for one PV only!*/
void dim_call_back();
int no_link = -1;

#define MAX_DIMRecords 1000
static int  nofDIMRecords = 0;
static double* DIMS_Data[MAX_DIMRecords];
static int*  DIMS_Records[MAX_DIMRecords];
static int  DIMS_ServiceID[MAX_DIMRecords];
 
/*DIM declarations */

/*Create the dset for devcaDIMServerSoft */
static long init_record();
static long read_caDIMServer();
static long write_caDIMServer();
struct {
	long		number;
	DEVSUPFUN	report;
	DEVSUPFUN	init;
	DEVSUPFUN	init_record;
	DEVSUPFUN	get_ioint_info;
	DEVSUPFUN	write_caDIMServer;
}devcaDIMServerSoft={
	5,
	NULL,
	NULL,
	init_record,
	NULL,
	write_caDIMServer,
};
epicsExportAddress(dset,devcaDIMServerSoft);


static long init_record(pcaDIMServer)
    struct caDIMServerRecord	*pcaDIMServer;
{
    char dim_name[1024];
    if(recGblInitConstantLink(&pcaDIMServer->inp,DBF_DOUBLE,&pcaDIMServer->val)){
         pcaDIMServer->udf = FALSE;
    }
    strcpy(dim_name,pcaDIMServer->dimservicename);
    if(nofDIMRecords<MAX_DIMRecords){
	/* alloc memeory for DIM data */
	DIMS_Data[nofDIMRecords]=calloc(1, sizeof(double));
	/* store pointer to Record */
	DIMS_Records[nofDIMRecords]=(int*) pcaDIMServer;
	/*register DIM service */
	DIMS_ServiceID[nofDIMRecords]=dis_add_service( dim_name, "D:1" , DIMS_Data[nofDIMRecords], sizeof(double), (void *)0, 0 );
	nofDIMRecords++;
	dis_start_serving(EPICS_DIM_SERVER);
    }
    return(0);
}

static long read_caDIMServer(pcaDIMServer)
    struct caDIMServerRecord	*pcaDIMServer;
{
    long status;

    status = dbGetLink(&(pcaDIMServer->inp),DBF_DOUBLE, &(pcaDIMServer->val),0,0);
    /*If return was succesful then set undefined false*/
    if(!status) pcaDIMServer->udf = FALSE;
    return(0);
}


/*code snip from the AppDevGuide*/
static long write_caDIMServer(pcaDIMServer)
    struct caDIMServerRecord *pcaDIMServer;
{

    long status;
    int ii=0;

    for(ii=0;ii<MAX_DIMRecords;ii++){
	if(DIMS_Records[ii]==(int*) pcaDIMServer) break;
    }
    /* copy DIM data buffer to record value */
    (double) *DIMS_Data[ii] = pcaDIMServer->val;
    /*printf("<i> devcaDIMServerSoft.c: write_caDIMServer: Recordnum %d : dim_data =%f \n", ii, pcaDIMServer->val);*/
    dis_update_service(DIMS_ServiceID[ii]);

    status=FALSE;
    if(status) {
	recGblSetSevr(pcaDIMServer,LINK_ALARM,INVALID_ALARM);
	return(status);
    }
    pcaDIMServer->udf = FALSE;
    pcaDIMServer->pact = FALSE;
    return(0);
}
